@@ -6,13 +6,14 @@ import java.math.BigInteger;
 import java.util.Arrays;
 
 import com.fasterxml.jackson.core.*;
-import com.fasterxml.jackson.core.JsonParser.Feature;
 import com.fasterxml.jackson.core.io.IOContext;
+import com.fasterxml.jackson.core.io.ContentReference;
 import com.fasterxml.jackson.core.io.NumberInput;
 import com.fasterxml.jackson.core.json.DupDetector;
 import com.fasterxml.jackson.core.json.JsonReadContext;
 import com.fasterxml.jackson.core.json.PackageVersion;
 import com.fasterxml.jackson.core.util.ByteArrayBuilder;
+import com.fasterxml.jackson.core.util.JacksonFeatureSet;
 import com.fasterxml.jackson.core.util.TextBuffer;
 
 /**
@@ -22,6 +23,11 @@ import com.fasterxml.jackson.core.util.TextBuffer;
  */
 public abstract class ParserBase extends ParserMinimalBase
 {
+    // JSON capabilities are the same as defaults
+    // @since 2.12
+    protected final static JacksonFeatureSet<StreamReadCapability> JSON_READ_CAPABILITIES
+        = DEFAULT_READ_CAPABILITIES;
+
     /*
     /**********************************************************
     /* Generic I/O state
@@ -350,9 +356,8 @@ public abstract class ParserBase extends ParserMinimalBase
         if (_currToken == JsonToken.START_OBJECT || _currToken == JsonToken.START_ARRAY) {
             ctxt = ctxt.getParent();
         }
-        /* 24-Sep-2013, tatu: Unfortunate, but since we did not expose exceptions,
-         *   need to wrap this here
-         */
+        // 24-Sep-2013, tatu: Unfortunate, but since we did not expose exceptions,
+        //   need to wrap this here
         try {
             ctxt.setCurrentName(name);
         } catch (IOException e) {
@@ -385,7 +390,7 @@ public abstract class ParserBase extends ParserMinimalBase
      */
     @Override
     public JsonLocation getTokenLocation() {
-        return new JsonLocation(_getSourceReference(),
+        return new JsonLocation(_contentReference(),
                 -1L, getTokenCharacterOffset(), // bytes, chars
                 getTokenLineNr(),
                 getTokenColumnNr());
@@ -398,7 +403,7 @@ public abstract class ParserBase extends ParserMinimalBase
     @Override
     public JsonLocation getCurrentLocation() {
         int col = _inputPtr - _currInputRowStart + 1; // 1-based
-        return new JsonLocation(_getSourceReference(),
+        return new JsonLocation(_contentReference(),
                 -1L, _currInputProcessed + _inputPtr, // bytes, chars
                 _currInputRow, col);
     }
@@ -464,6 +469,9 @@ public abstract class ParserBase extends ParserMinimalBase
      * reader. This may be called along with {@link #_closeInput} (for
      * example, when explicitly closing this reader instance), or
      * separately (if need be).
+     *
+     * @throws IOException Not thrown by base implementation but could be thrown
+     *   by sub-classes
      */
     protected void _releaseBuffers() throws IOException {
         _textBuffer.releaseBuffers();
@@ -486,12 +494,17 @@ public abstract class ParserBase extends ParserMinimalBase
             _reportInvalidEOF(String.format(
                     ": expected close marker for %s (start marker at %s)",
                     marker,
-                    _parsingContext.getStartLocation(_getSourceReference())),
+                    _parsingContext.startLocation(_contentReference())),
                     null);
         }
     }
 
     /**
+     * @return If no exception is thrown, {@code -1} which is used as marked for "end-of-input"
+     *
+     * @throws JsonParseException If check on {@code _handleEOF()} fails; usually because
+     *    the current context is not root context (missing end markers in content)
+     *
      * @since 2.4
      */
     protected final int _eofAsNextChar() throws JsonParseException {
@@ -576,7 +589,7 @@ public abstract class ParserBase extends ParserMinimalBase
     /* Numeric accessors of public API
     /**********************************************************
      */
-    
+
     @Override
     public Number getNumberValue() throws IOException
     {
@@ -594,13 +607,11 @@ public abstract class ParserBase extends ParserMinimalBase
             if ((_numTypesValid & NR_BIGINT) != 0) {
                 return _numberBigInt;
             }
-            // Shouldn't get this far but if we do
-            return _numberBigDecimal;
+            _throwInternal();
         }
-    
-        /* And then floating point types. But here optimal type
-         * needs to be big decimal, to avoid losing any data?
-         */
+
+        // And then floating point types. But here optimal type
+        // needs to be big decimal, to avoid losing any data?
         if ((_numTypesValid & NR_BIGDECIMAL) != 0) {
             return _numberBigDecimal;
         }
@@ -609,7 +620,39 @@ public abstract class ParserBase extends ParserMinimalBase
         }
         return _numberDouble;
     }
-    
+
+    // NOTE: mostly copied from above
+    @Override
+    public Number getNumberValueExact() throws IOException
+    {
+        if (_currToken == JsonToken.VALUE_NUMBER_INT) {
+            if (_numTypesValid == NR_UNKNOWN) {
+                _parseNumericValue(NR_UNKNOWN);
+            }
+            if ((_numTypesValid & NR_INT) != 0) {
+                return _numberInt;
+            }
+            if ((_numTypesValid & NR_LONG) != 0) {
+                return _numberLong;
+            }
+            if ((_numTypesValid & NR_BIGINT) != 0) {
+                return _numberBigInt;
+            }
+            _throwInternal();
+        }
+        // 09-Jul-2020, tatu: [databind#2644] requires we will retain accuracy, so:
+        if (_numTypesValid == NR_UNKNOWN) {
+            _parseNumericValue(NR_BIGDECIMAL);
+        }
+        if ((_numTypesValid & NR_BIGDECIMAL) != 0) {
+            return _numberBigDecimal;
+        }
+        if ((_numTypesValid & NR_DOUBLE) == 0) { // sanity check
+            _throwInternal();
+        }
+        return _numberDouble;
+    }
+
     @Override
     public NumberType getNumberType() throws IOException
     {
@@ -728,7 +771,7 @@ public abstract class ParserBase extends ParserMinimalBase
     /* Conversion from textual to numeric representation
     /**********************************************************
      */
-    
+
     /**
      * Method that will parse actual numeric value out of a syntactically
      * valid number value. Type it will parse into depends on whether
@@ -737,14 +780,24 @@ public abstract class ParserBase extends ParserMinimalBase
      *
      * @param expType Numeric type that we will immediately need, if any;
      *   mostly necessary to optimize handling of floating point numbers
+     *
+     * @throws IOException If there are problems reading content
+     * @throws JsonParseException If there are problems decoding number value
      */
     protected void _parseNumericValue(int expType) throws IOException
     {
+        // 12-Jun-2020, tatu: Sanity check to prevent more cryptic error for this case.
+        //    (note: could alternatively see if TextBuffer has aggregated contents, avoid
+        //    exception -- but that might be more confusing)
+        if (_closed) {
+            _reportError("Internal error: _parseNumericValue called when parser instance closed");
+        }
+
         // Int or float?
         if (_currToken == JsonToken.VALUE_NUMBER_INT) {
-            int len = _intLength;
+            final int len = _intLength;
             // First: optimization for simple int
-            if (len <= 9) { 
+            if (len <= 9) {
                 int i = _textBuffer.contentsAsInt(_numberNegative);
                 _numberInt = i;
                 _numTypesValid = NR_INT;
@@ -782,11 +835,15 @@ public abstract class ParserBase extends ParserMinimalBase
         _reportError("Current token (%s) not numeric, can not use numeric value accessors", _currToken);
     }
 
-    /**
-     * @since 2.6
-     */
+    // @since 2.6
     protected int _parseIntValue() throws IOException
     {
+        // 12-Jun-2020, tatu: Sanity check to prevent more cryptic error for this case.
+        //    (note: could alternatively see if TextBuffer has aggregated contents, avoid
+        //    exception -- but that might be more confusing)
+        if (_closed) {
+            _reportError("Internal error: _parseNumericValue called when parser instance closed");
+        }
         // Inlined variant of: _parseNumericValue(NR_INT)
         if (_currToken == JsonToken.VALUE_NUMBER_INT) {
             if (_intLength <= 9) {
@@ -1015,7 +1072,8 @@ public abstract class ParserBase extends ParserMinimalBase
         JsonReadContext ctxt = getParsingContext();
         _reportError(String.format(
                 "Unexpected close marker '%s': expected '%c' (for %s starting at %s)",
-                (char) actCh, expCh, ctxt.typeDesc(), ctxt.getStartLocation(_getSourceReference())));
+                (char) actCh, expCh, ctxt.typeDesc(),
+                ctxt.startLocation(_contentReference())));
     }
 
     @SuppressWarnings("deprecation")
@@ -1035,7 +1093,13 @@ public abstract class ParserBase extends ParserMinimalBase
     /**
      * Method called to report a problem with unquoted control character.
      * Note: it is possible to suppress some instances of
-     * exception by enabling {@link Feature#ALLOW_UNQUOTED_CONTROL_CHARS}.
+     * exception by enabling
+     * {@link com.fasterxml.jackson.core.json.JsonReadFeature#ALLOW_UNESCAPED_CONTROL_CHARS}.
+     *
+     * @param i Invalid control character
+     * @param ctxtDesc Addition description of context to use in exception message
+     *
+     * @throws JsonParseException explaining the problem
      */
     @SuppressWarnings("deprecation")
     protected void _throwUnquotedSpace(int i, String ctxtDesc) throws JsonParseException {
@@ -1052,6 +1116,8 @@ public abstract class ParserBase extends ParserMinimalBase
      *    invalid (unrecognized) JSON token: called when parser finds something that
      *    looks like unquoted textual token
      *
+     * @throws IOException Not thrown by base implementation but allowed by sub-classes
+     *
      * @since 2.10
      */
     protected String _validJsonTokenList() throws IOException {
@@ -1063,6 +1129,8 @@ public abstract class ParserBase extends ParserMinimalBase
      *    invalid (unrecognized) JSON value: called when parser finds something that
      *    does not look like a value or separator.
      *
+     * @throws IOException Not thrown by base implementation but allowed by sub-classes
+     *
      * @since 2.10
      */
     @SuppressWarnings("deprecation")
@@ -1083,6 +1151,10 @@ public abstract class ParserBase extends ParserMinimalBase
      * Method that sub-classes must implement to support escaped sequences
      * in base64-encoded sections.
      * Sub-classes that do not need base64 support can leave this as is
+     *
+     * @return Character decoded, if any
+     *
+     * @throws IOException If escape decoding fails
      */
     protected char _decodeEscaped() throws IOException {
         throw new UnsupportedOperationException();
@@ -1138,9 +1210,9 @@ public abstract class ParserBase extends ParserMinimalBase
         return reportInvalidBase64Char(b64variant, ch, bindex, null);
     }
 
-    /**
+    /*
      * @param bindex Relative index within base64 character unit; between 0
-     *   and 3 (as unit has exactly 4 characters)
+     *  and 3 (as unit has exactly 4 characters)
      */
     protected IllegalArgumentException reportInvalidBase64Char(Base64Variant b64variant, int ch, int bindex, String msg) throws IllegalArgumentException {
         String base;
@@ -1174,18 +1246,33 @@ public abstract class ParserBase extends ParserMinimalBase
      */
 
     /**
-     * Helper method used to encapsulate logic of including (or not) of
-     * "source reference" when constructing {@link JsonLocation} instances.
-     *
+     * @return Source reference
      * @since 2.9
+     * @deprecated Since 2.13, use {@link #_contentReference()} instead.
      */
+    @Deprecated
     protected Object _getSourceReference() {
         if (JsonParser.Feature.INCLUDE_SOURCE_IN_LOCATION.enabledIn(_features)) {
-            return _ioContext.getSourceReference();
+            return _ioContext.contentReference().getRawContent();
         }
         return null;
     }
 
+    /**
+     * Helper method used to encapsulate logic of including (or not) of
+     * "content reference" when constructing {@link JsonLocation} instances.
+     *
+     * @return Source reference object, if any; {@code null} if none
+     *
+     * @since 2.13
+     */
+    protected ContentReference _contentReference() {
+        if (JsonParser.Feature.INCLUDE_SOURCE_IN_LOCATION.enabledIn(_features)) {
+            return _ioContext.contentReference();
+        }
+        return ContentReference.unknown();
+    }
+
     protected static int[] growArrayBy(int[] arr, int more)
     {
         if (arr == null) {
