@@ -1,6 +1,7 @@
 package com.fasterxml.jackson.core.read;
 
 import java.io.ByteArrayInputStream;
+import java.io.CharArrayReader;
 import java.io.IOException;
 import java.io.StringReader;
 import java.math.BigDecimal;
@@ -8,6 +9,7 @@ import java.math.BigInteger;
 
 import com.fasterxml.jackson.core.*;
 import com.fasterxml.jackson.core.exc.InputCoercionException;
+import com.fasterxml.jackson.core.io.NumberInput;
 import com.fasterxml.jackson.core.json.JsonReadFeature;
 
 /**
@@ -18,7 +20,13 @@ import com.fasterxml.jackson.core.json.JsonReadFeature;
 public class NumberParsingTest
     extends com.fasterxml.jackson.core.BaseTest
 {
-    private final JsonFactory FACTORY = new JsonFactory();
+    private final JsonFactory FACTORY = sharedStreamFactory();
+
+    /*
+    /**********************************************************************
+    /* Tests, Boolean
+    /**********************************************************************
+     */
 
     public void testSimpleBoolean() throws Exception
     {
@@ -38,9 +46,15 @@ public class NumberParsingTest
         p.close();
     }
 
+    /*
+    /**********************************************************************
+    /* Tests, Int
+    /**********************************************************************
+     */
+
     public void testSimpleInt() throws Exception
     {
-        for (int EXP_I : new int[] { 1234, -999, 0, 1, -2 }) {
+        for (int EXP_I : new int[] { 1234, -999, 0, 1, -2, 123456789 }) {
             _testSimpleInt(EXP_I, MODE_INPUT_STREAM);
             _testSimpleInt(EXP_I, MODE_INPUT_STREAM_THROTTLED);
             _testSimpleInt(EXP_I, MODE_READER);
@@ -55,8 +69,34 @@ public class NumberParsingTest
         assertToken(JsonToken.START_ARRAY, p.nextToken());
         assertToken(JsonToken.VALUE_NUMBER_INT, p.nextToken());
         assertEquals(JsonParser.NumberType.INT, p.getNumberType());
+        assertTrue(p.isExpectedNumberIntToken());
         assertEquals(""+EXP_I, p.getText());
 
+        if (((short) EXP_I) == EXP_I) {
+            assertEquals((short) EXP_I, p.getShortValue());
+            if (((byte) EXP_I) == EXP_I) {
+                assertEquals((byte) EXP_I, p.getByteValue());
+            } else {
+                // verify overflow
+                try {
+                    p.getByteValue();
+                    fail("Should get exception for non-byte value "+EXP_I);
+                } catch (InputCoercionException e) {
+                    verifyException(e, "Numeric value");
+                    verifyException(e, "out of range");
+                }
+            }
+        } else {
+            // verify overflow
+            try {
+                p.getShortValue();
+                fail("Should get exception for non-short value "+EXP_I);
+            } catch (InputCoercionException e) {
+                verifyException(e, "Numeric value");
+                verifyException(e, "out of range");
+            }
+        }
+
         assertEquals(EXP_I, p.getIntValue());
         assertEquals(EXP_I, p.getValueAsInt(EXP_I + 3));
         assertEquals(EXP_I, p.getValueAsInt());
@@ -69,20 +109,29 @@ public class NumberParsingTest
         DOC = String.valueOf(EXP_I);
         p = createParser(mode, DOC + " "); // DataInput requires separator
         assertToken(JsonToken.VALUE_NUMBER_INT, p.nextToken());
+        assertTrue(p.isExpectedNumberIntToken());
         assertEquals(DOC, p.getText());
 
-        int i = 0;
-        
-        try {
-            i = p.getIntValue();
-        } catch (Exception e) {
-            throw new Exception("Failed to parse input '"+DOC+"' (parser of type "+p.getClass().getSimpleName()+")", e);
-        }
+        int i = p.getIntValue();
+
         assertEquals(EXP_I, i);
         assertEquals((long) EXP_I, p.getLongValue());
         assertEquals((double) EXP_I, p.getDoubleValue());
         assertEquals(BigDecimal.valueOf((long) EXP_I), p.getDecimalValue());
         p.close();
+
+        // and finally, coercion from double to int; couple of variants
+        DOC = String.valueOf(EXP_I)+".0";
+        p = createParser(mode, DOC + " ");
+        assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+        assertFalse(p.isExpectedNumberIntToken());
+        assertEquals(EXP_I, p.getValueAsInt());
+        p.close();
+
+        p = createParser(mode, DOC + " ");
+        assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+        assertEquals(EXP_I, p.getValueAsInt(0));
+        p.close();
     }
 
     public void testIntRange() throws Exception
@@ -103,6 +152,68 @@ public class NumberParsingTest
         }
     }
 
+    public void testIntParsing() throws Exception
+    {
+        char[] testChars = "123456789".toCharArray();
+
+        assertEquals(3, NumberInput.parseInt(testChars, 2, 1));
+        assertEquals(123, NumberInput.parseInt(testChars, 0, 3));
+        assertEquals(2345, NumberInput.parseInt(testChars, 1, 4));
+        assertEquals(9, NumberInput.parseInt(testChars, 8, 1));
+        assertEquals(456789, NumberInput.parseInt(testChars, 3, 6));
+        assertEquals(23456, NumberInput.parseInt(testChars, 1, 5));
+        assertEquals(123456789, NumberInput.parseInt(testChars, 0, 9));
+
+        testChars = "32".toCharArray();
+        assertEquals(32, NumberInput.parseInt(testChars, 0, 2));
+        testChars = "189".toCharArray();
+        assertEquals(189, NumberInput.parseInt(testChars, 0, 3));
+
+        testChars = "10".toCharArray();
+        assertEquals(10, NumberInput.parseInt(testChars, 0, 2));
+        assertEquals(0, NumberInput.parseInt(testChars, 1, 1));
+    }
+
+    public void testIntParsingWithStrings() throws Exception
+    {
+        assertEquals(3, NumberInput.parseInt("3"));
+        assertEquals(3, NumberInput.parseInt("+3"));
+        assertEquals(0, NumberInput.parseInt("0"));
+        assertEquals(-3, NumberInput.parseInt("-3"));
+        assertEquals(27, NumberInput.parseInt("27"));
+        assertEquals(-31, NumberInput.parseInt("-31"));
+        assertEquals(271, NumberInput.parseInt("271"));
+        assertEquals(-131, NumberInput.parseInt("-131"));
+        assertEquals(2709, NumberInput.parseInt("2709"));
+        assertEquals(-9999, NumberInput.parseInt("-9999"));
+        assertEquals(Integer.MIN_VALUE, NumberInput.parseInt(""+Integer.MIN_VALUE));
+        assertEquals(Integer.MAX_VALUE, NumberInput.parseInt(""+Integer.MAX_VALUE));
+    }
+
+    public void testLongParsingWithStrings() throws Exception
+    {
+        assertEquals(3, NumberInput.parseLong("3"));
+        assertEquals(3, NumberInput.parseLong("+3"));
+        assertEquals(0, NumberInput.parseLong("0"));
+        assertEquals(-3, NumberInput.parseLong("-3"));
+        assertEquals(27, NumberInput.parseLong("27"));
+        assertEquals(-31, NumberInput.parseLong("-31"));
+        assertEquals(271, NumberInput.parseLong("271"));
+        assertEquals(-131, NumberInput.parseLong("-131"));
+        assertEquals(2709, NumberInput.parseLong("2709"));
+        assertEquals(-9999, NumberInput.parseLong("-9999"));
+        assertEquals(Long.MIN_VALUE, NumberInput.parseLong(""+Long.MIN_VALUE));
+        assertEquals(Integer.MIN_VALUE-1, NumberInput.parseLong(""+(Integer.MIN_VALUE-1)));
+        assertEquals(Long.MAX_VALUE, NumberInput.parseLong(""+Long.MAX_VALUE));
+        assertEquals(Integer.MAX_VALUE+1, NumberInput.parseLong(""+(Integer.MAX_VALUE+1)));
+    }
+
+    /*
+    /**********************************************************************
+    /* Tests, Long
+    /**********************************************************************
+     */
+
     public void testSimpleLong() throws Exception
     {
         _testSimpleLong(MODE_INPUT_STREAM);
@@ -166,6 +277,55 @@ public class NumberParsingTest
         }
     }
 
+    public void testLongParsing() throws Exception
+    {
+        char[] testChars = "123456789012345678".toCharArray();
+
+        assertEquals(123456789012345678L, NumberInput.parseLong(testChars, 0, testChars.length));
+    }
+
+    public void testLongBoundsChecks() throws Exception
+    {
+        String minLong = String.valueOf(Long.MIN_VALUE).substring(1);
+        String belowMinLong = BigInteger.valueOf(Long.MIN_VALUE).subtract(BigInteger.ONE)
+                .toString().substring(1);
+        String maxLong = String.valueOf(Long.MAX_VALUE);
+        String aboveMaxLong = BigInteger.valueOf(Long.MAX_VALUE).add(BigInteger.ONE).toString();
+        final String VALUE_491 = "1323372036854775807"; // is within range (JACKSON-491)
+        final String OVERFLOW =  "9999999999999999999"; // and this one is clearly out
+
+        assertTrue(NumberInput.inLongRange(minLong, true));
+        assertTrue(NumberInput.inLongRange(maxLong, false));
+        assertTrue(NumberInput.inLongRange(VALUE_491, true));
+        assertTrue(NumberInput.inLongRange(VALUE_491, false));
+
+        assertFalse(NumberInput.inLongRange(OVERFLOW, false));
+        assertFalse(NumberInput.inLongRange(OVERFLOW, true));
+        assertFalse(NumberInput.inLongRange(belowMinLong, true));
+        assertFalse(NumberInput.inLongRange(aboveMaxLong, false));
+
+        char[] cbuf = minLong.toCharArray();
+        assertTrue(NumberInput.inLongRange(cbuf, 0, cbuf.length, true));
+        cbuf = maxLong.toCharArray();
+        assertTrue(NumberInput.inLongRange(cbuf, 0, cbuf.length, false));
+        cbuf = VALUE_491.toCharArray();
+        assertTrue(NumberInput.inLongRange(cbuf, 0, cbuf.length, true));
+        assertTrue(NumberInput.inLongRange(cbuf, 0, cbuf.length, false));
+
+        cbuf = OVERFLOW.toCharArray();
+        assertFalse(NumberInput.inLongRange(cbuf, 0, cbuf.length, true));
+        assertFalse(NumberInput.inLongRange(cbuf, 0, cbuf.length, false));
+
+        cbuf = aboveMaxLong.toCharArray();
+        assertFalse(NumberInput.inLongRange(cbuf, 0, cbuf.length, false));
+    }
+    
+    /*
+    /**********************************************************************
+    /* Tests, BigXxx
+    /**********************************************************************
+     */
+    
     public void testBigDecimalRange() throws Exception
     {
         for (int mode : ALL_MODES) {
@@ -207,6 +367,12 @@ public class NumberParsingTest
             p.close();
         }
     }
+
+    /*
+    /**********************************************************************
+    /* Tests, floating point (basic)
+    /**********************************************************************
+     */
     
     public void testSimpleDouble() throws Exception
     {
@@ -256,6 +422,113 @@ public class NumberParsingTest
         }
     }
 
+    public void testFloatBoundary146Chars() throws Exception
+    {
+        final char[] arr = new char[50005];
+        for(int i = 500; i != 9000; ++i) {
+          java.util.Arrays.fill(arr, 0, i, ' ');
+          arr[i] = '-';
+          arr[i + 1] = '1';
+          arr[i + 2] = 'e';
+          arr[i + 3] = '-';
+          arr[i + 4] = '1';
+          CharArrayReader r = new CharArrayReader(arr, 0, i+5);
+          JsonParser p = FACTORY.createParser(r);
+          assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+          p.close();
+        }
+    }
+
+    public void testFloatBoundary146Bytes() throws Exception
+    {
+        final byte[] arr = new byte[50005];
+        for(int i = 500; i != 9000; ++i) {
+            java.util.Arrays.fill(arr, 0, i, (byte) 0x20);
+            arr[i] = '-';
+            arr[i + 1] = '1';
+            arr[i + 2] = 'e';
+            arr[i + 3] = '-';
+            arr[i + 4] = '1';
+            ByteArrayInputStream in = new ByteArrayInputStream(arr, 0, i+5);
+            JsonParser p = FACTORY.createParser(in);
+            assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+            p.close();
+        }
+    }
+
+    /*
+    /**********************************************************************
+    /* Tests, BigDecimal (core#677)
+    /**********************************************************************
+     */
+
+    public void testBigBigDecimalsBytes() throws Exception
+    {
+        _testBigBigDecimals(MODE_INPUT_STREAM);
+        _testBigBigDecimals(MODE_INPUT_STREAM_THROTTLED);
+    }
+
+    public void testBigBigDecimalsChars() throws Exception
+    {
+        _testBigBigDecimals(MODE_READER);
+    }
+
+    public void testBigBigDecimalsDataInput() throws Exception
+    {
+        _testBigBigDecimals(MODE_DATA_INPUT);
+    }
+
+    private void _testBigBigDecimals(int mode) throws Exception
+    {
+        final String BASE_FRACTION =
+ "01610253934481930774151441507943554511027782188707463024288149352877602369090537"
++"80583522838238149455840874862907649203136651528841378405339370751798532555965157588"
++"51877960056849468879933122908090021571162427934915567330612627267701300492535817858"
++"36107216979078343419634586362681098115326893982589327952357032253344676618872460059"
++"52652865429180458503533715200184512956356092484787210672008123556320998027133021328"
++"04777044107393832707173313768807959788098545050700242134577863569636367439867566923"
++"33479277494056927358573496400831024501058434838492057410330673302052539013639792877"
++"76670882022964335417061758860066263335250076803973514053909274208258510365484745192"
++"39425298649420795296781692303253055152441850691276044546565109657012938963181532017"
++"97420631515930595954388119123373317973532146157980827838377034575940814574561703270"
++"54949003909864767732479812702835339599792873405133989441135669998398892907338968744"
++"39682249327621463735375868408190435590094166575473967368412983975580104741004390308"
++"45302302121462601506802738854576700366634229106405188353120298347642313881766673834"
++"60332729485083952142460470270121052469394888775064758246516888122459628160867190501"
++"92476878886543996441778751825677213412487177484703116405390741627076678284295993334"
++"23142914551517616580884277651528729927553693274406612634848943914370188078452131231"
++"17351787166509190240927234853143290940647041705485514683182501795615082930770566118"
++"77488417962195965319219352314664764649802231780262169742484818333055713291103286608"
++"64318433253572997833038335632174050981747563310524775762280529871176578487487324067"
++"90242862159403953039896125568657481354509805409457993946220531587293505986329150608"
++"18702520420240989908678141379300904169936776618861221839938283876222332124814830207"
++"073816864076428273177778788053613345444299361357958409716099682468768353446625063";
+
+        for (String asText : new String[] {
+                "50."+BASE_FRACTION,
+                "-37."+BASE_FRACTION,
+                "0.00"+BASE_FRACTION,
+                "-0.012"+BASE_FRACTION,
+                "9999998."+BASE_FRACTION,
+                "-8888392."+BASE_FRACTION,
+        }) {
+            final String DOC = "[ "+asText+" ]";
+
+            JsonParser p = createParser(mode, DOC);
+            assertToken(JsonToken.START_ARRAY, p.nextToken());
+            assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+            final BigDecimal exp = new BigDecimal(asText);
+            assertEquals(exp, p.getDecimalValue());
+            p.close();
+        }
+    }
+
+    /*
+    /**********************************************************************
+    /* Tests, misc other
+    /**********************************************************************
+     */
+
     public void testNumbers() throws Exception
     {
         _testNumbers(MODE_INPUT_STREAM);
